from __future__ import annotations

import uuid
from pathlib import Path

from pytest import raises
from sqlalchemy.orm import Session

from . import repository

UUID_STR = "4f80f02f-52e3-4fe2-b9f2-2c3e99449ce9"
UUID = uuid.UUID(UUID_STR)


def test_rel_path(session: Session):
    p = repository.rel_path(UUID)
    expected = Path("4f", "80", "4f80f02f-52e3-4fe2-b9f2-2c3e99449ce9")
    assert isinstance(p, Path)
    assert p == expected


def test_abs_path(session: Session):
    p = repository.abs_path(UUID)
    assert isinstance(p, Path)

    # FIXME: fails on a Mac due to symlinks /var@ -> /private/var
    # expected = Path(self.app.instance_path, 'data', 'files',
    #                 '4f', '80', '4f80f02f-52e3-4fe2-b9f2-2c3e99449ce9')
    # self.assertEquals(p, expected)


def test_get_with_some_content(session: Session):
    p = repository.abs_path(UUID)
    if not p.parent.exists():
        p.parent.mkdir(parents=True)
    p.open("wb").write(b"my file content")

    val = repository.get(UUID)
    assert val == p
    assert val.open("rb").read() == b"my file content"

    # __getitem__
    val = repository[UUID]
    assert val == p
    assert val.open("rb").read() == b"my file content"


def test_get_with_non_existing_content(session: Session):
    # non-existent
    u = uuid.UUID("bcdc32ac-498d-4544-9e7f-fb2c75097011")
    default_path = Path("/tmp/default-path")
    assert repository.get(u) is None
    assert repository.get(u, default=default_path) is default_path

    # __getitem__ non-existent
    with raises(KeyError):
        assert repository[u]


def test_set(session: Session):
    u1 = uuid.uuid4()
    p = repository.abs_path(u1)
    repository.set(u1, b"my file content")
    assert p.open("rb").read() == b"my file content"


def test_setitem(session: Session):
    u1 = uuid.uuid4()
    p = repository.abs_path(u1)
    repository[u1] = b"my file content"
    assert p.open("rb").read() == b"my file content"
    # FIXME: test Unicode content


def test_delete(session: Session):
    u1 = uuid.uuid4()
    repository.set(u1, b"my file content")
    p = repository.abs_path(u1)
    assert p.exists()

    repository.delete(u1)
    assert not p.exists()


def test_delitem(session: Session):
    u1 = uuid.uuid4()
    repository.set(u1, b"my file content")
    p = repository.abs_path(u1)
    assert p.exists()

    del repository[u1]
    assert not p.exists()


def test_delete_non_existent(session: Session):
    # non-existent
    u1 = uuid.uuid4()
    with raises(KeyError):
        repository.delete(u1)

    # same w/ __delitem__
    with raises(KeyError):
        del repository[u1]



from __future__ import annotations

import logging
import os
from contextlib import contextmanager
from pathlib import Path
from tempfile import mkstemp
from typing import Iterator, Optional

logger = logging.getLogger(__name__)

# Hack for Mac OS + homebrew
os.environ["PATH"] += ":/usr/local/bin"

TMP_DIR = "tmp"
CACHE_DIR = "cache"


def get_tmp_dir() -> Path:
    from . import converter

    return converter.tmp_dir


# Utils
@contextmanager
def make_temp_file(
    blob: bytes | None = None,
    prefix: str = "tmp",
    suffix: str = "",
    tmp_dir: Path | None = None,
) -> Iterator[str]:

    if tmp_dir is None:
        tmp_dir = get_tmp_dir()

    fd, filename = mkstemp(dir=str(tmp_dir), prefix=prefix, suffix=suffix)
    if blob is not None:
        io = os.fdopen(fd, "wb")
        io.write(blob)
        io.close()
    else:
        os.close(fd)

    yield filename

    try:
        os.remove(filename)
    except OSError:
        pass



""""""
from __future__ import annotations

from pytest import fixture

from abilian.core.entities import Entity
from abilian.core.models.subjects import User

from .service import ActivityService


class Message1(Entity):
    pass


@fixture
def activity_service(app, db):
    service = ActivityService()
    service.start()
    yield service
    service.stop()


def test(app, session, activity_service):
    service = activity_service
    user = User(email="test@example.com")
    message = Message1(creator=user, owner=user, name="test message")

    with session.begin_nested():
        session.add(user)
        session.add(message)
    service.log_activity(None, user, "post", message)
    session.flush()

    m_id, m_type = message.id, message.entity_type

    entries = service.entries_for_actor(user, 10)
    assert len(entries) == 1

    entry = entries[0]
    assert entry.actor == user
    assert entry.actor_id == user.id
    assert entry.object == message
    assert entry.object_type == m_type
    assert entry._fk_object_id == m_id
    assert entry.object_id == m_id
    assert entry.target is None

    # test entry doesn't reference target if its in deleted state
    session.delete(message)
    session.flush()
    # send activity after flush: activity should not reference
    # an instance in "deleted state"
    service.log_activity(None, user, "delete", message)
    session.flush()

    entries = service.entries_for_actor(user, 10)
    assert len(entries) == 2

    entry = entries[1]
    assert entry.object is None
    assert entry.object_type == m_type
    assert entry._fk_object_id is None
    assert entry.object_id == m_id
    assert entry.target is None



from __future__ import annotations

import logging
import typing
from datetime import datetime, timedelta
from typing import Any, Callable, Dict, List, Optional

from flask import Flask, current_app, g, redirect, request, url_for
from flask_babel import lazy_gettext as _l
from flask_login import current_user, login_user, user_logged_in, user_logged_out
from werkzeug.exceptions import Forbidden
from werkzeug.wrappers import Response

from abilian.core.extensions import db, login_manager
from abilian.core.models.subjects import User
from abilian.core.signals import user_loaded
from abilian.core.util import unwrap
from abilian.services import Service, ServiceState
from abilian.services.security.service import SecurityService
from abilian.web.action import DynamicIcon, actions
from abilian.web.nav import NavGroup, NavItem

from .models import LoginSession
from .views import login as login_views

if typing.TYPE_CHECKING:
    from abilian.app import Application

__all__ = ["AuthService", "user_menu"]

logger = logging.getLogger(__name__)


def is_anonymous(context: dict[str, bool]) -> bool:
    return current_user.is_anonymous


def is_authenticated(context: dict[str, bool]) -> bool:
    return not is_anonymous(context)


def _user_photo_endpoint() -> str:
    from abilian.web.views import images  # lazy import: avoid circular import

    return images.user_url_args(current_user, 16)[0]


def _user_photo_icon_args(icon: DynamicIcon, url_args: dict) -> dict[str, Any]:
    from abilian.web.views import images  # lazy import avoid circular import

    return images.user_url_args(current_user, max(icon.width, icon.height))[1]


user_menu = NavGroup(
    "user",
    "authenticated",
    title=lambda c: current_user.name,
    icon=DynamicIcon(
        endpoint=_user_photo_endpoint,
        css="avatar",
        size=20,
        url_args=_user_photo_icon_args,
    ),
    condition=is_authenticated,
    items=(
        NavItem(
            "user",
            "logout",
            title=_l("Logout"),
            icon="log-out",
            url=lambda context: url_for("login.logout"),
            divider=True,
        ),
    ),
)

_ACTIONS = (
    NavItem(
        "user",
        "login",
        title=_l("Login"),
        icon="log-in",
        url=lambda context: url_for("login.login_form"),
        condition=is_anonymous,
    ),
    user_menu,
)


class AuthServiceState(ServiceState):
    """State class for :class:`AuthService`"""

    def __init__(self, service: AuthService, *args: Any, **kwargs: Any):
        super().__init__(service, *args, **kwargs)
        self.bp_access_controllers: dict[str | None, list[Callable]] = {None: []}
        self.endpoint_access_controllers: dict[str, list[Callable]] = {}

    def add_bp_access_controller(self, blueprint: str, func: Callable):
        self.bp_access_controllers.setdefault(blueprint, []).append(func)

    def add_endpoint_access_controller(self, endpoint: str, func: Callable):
        self.endpoint_access_controllers.setdefault(endpoint, []).append(func)


class AuthService(Service):
    name = "auth"
    AppStateClass = AuthServiceState

    def init_app(self, app: Flask):
        login_manager.init_app(app)
        login_manager.login_view = "login.login_form"

        super().init_app(app)

        self.login_url_prefix = app.config.get("LOGIN_URL", "/user")

        app.before_request(self.do_access_control)
        app.before_request(self.update_user_session_data)
        user_logged_in.connect(self.user_logged_in, sender=app)
        user_logged_out.connect(self.user_logged_out, sender=app)

        app.register_blueprint(login_views, url_prefix=self.login_url_prefix)

        with app.app_context():
            actions.register(*_ACTIONS)

    @login_manager.user_loader
    def load_user(user_id: str) -> User | None:
        try:
            user = User.query.get(user_id)

            if not user or not user.can_login:
                # if a user is edited and should not have access any more, this
                # will ensure they cannot continue if he had an active session
                return None
        except Exception:
            logger.warning("Error during login.", exc_info=True)
            session = db.session()
            if not session.is_active:
                # session is not usable, rollback should restore a usable
                # session
                session.rollback()

            return None

        app = unwrap(current_app)
        app.services[AuthService.name].user_logged_in(app, user)
        user_loaded.send(app, user=user)
        return user

    def user_logged_in(self, app: Application, user: User):
        # `g.user` is used as `current_user`, but `current_user` is actually looking
        # for `request.user` whereas `g` is on app local stack.
        #
        # `g.logged_user` is the actual user. In the future for example we may allow
        # a manager to see site as another user (impersonate), or propose a "see as
        # anonymous" function
        from abilian.services import get_security_service

        g.user = g.logged_user = user
        is_anonymous = user is None or user.is_anonymous
        security = get_security_service()
        g.is_manager = (
            user
            and not is_anonymous
            and (security.has_role(user, "admin") or security.has_role(user, "manager"))
        )

    def user_logged_out(self, app: Application, user: User):
        if hasattr(g, "user"):
            del g.user
            del g.logged_user
            del g.is_manager

    def redirect_to_login(self, next_url=True) -> Response:
        kw = {}
        if next_url is not False:
            kw["next"] = request.url if next_url is True else next_url

        return redirect(url_for(login_manager.login_view, **kw))

    def do_access_control(self) -> Response | None:
        """`before_request` handler to check if user should be redirected to
        login page."""
        from abilian.services import get_service

        if current_app.testing and current_app.config.get("NO_LOGIN"):
            # Special case for tests
            user = User.query.get(0)
            login_user(user, force=True)
            return None

        state = self.app_state
        user = unwrap(current_user)

        # Another special case for tests
        if current_app.testing and getattr(user, "is_admin", False):
            return None

        security: SecurityService = get_service("security")
        user_roles = frozenset(security.get_roles(user))
        endpoint = request.endpoint
        blueprint = request.blueprint

        access_controllers: list[Callable] = []
        access_controllers.extend(state.bp_access_controllers.get(None, []))

        if blueprint and blueprint in state.bp_access_controllers:
            access_controllers.extend(state.bp_access_controllers[blueprint])

        if endpoint and endpoint in state.endpoint_access_controllers:
            access_controllers.extend(state.endpoint_access_controllers[endpoint])

        for access_controller in reversed(access_controllers):
            verdict = access_controller(user=user, roles=user_roles)
            if verdict is None:
                continue
            elif verdict is True:
                return None
            else:
                if user.is_anonymous:
                    return self.redirect_to_login()
                raise Forbidden()

        # default policy
        if current_app.config.get("PRIVATE_SITE") and user.is_anonymous:
            return self.redirect_to_login()

        return None

    def update_user_session_data(self):
        user = current_user
        if current_user.is_anonymous:
            return

        # Update last_active every 60 seconds only so as to not stress
        # the database too much.
        now = datetime.utcnow()
        if user.last_active is None or (now - user.last_active) > timedelta(minutes=1):
            user.last_active = now
            db.session.add(user)
            db.session.commit()

        refresh_login_session(user)


def refresh_login_session(user: User):
    now = datetime.utcnow()
    session = LoginSession.query.get_active_for(user)
    if not session:
        return

    from_now = now - session.last_active_at

    if from_now > timedelta(hours=1):
        session.ended_at = session.last_active_at
        session = LoginSession.new()
        db.session.add(session)
        db.session.commit()

    elif from_now > timedelta(minutes=1):
        session.last_active_at = now
        db.session.commit()



"""Another try at an activity service.

See: ActivityStreams, OpenSocial

Note: ICOM also a notion of activity, but it's very different.
"""


from __future__ import annotations


class Activity:
    """Persistent object that represents an activity entry."""


class ActivityService:
    def post(self, activity):
        """Post an activity to the service."""

    # Now we need some methods to query the stream



"""Field filters for WTForm."""
from __future__ import annotations

from typing import Union

__all__ = ["strip", "uppercase", "lowercase"]


def strip(data: None | int | str) -> int | str:
    """Strip data if data is a string."""
    if data is None:
        return ""
    if not isinstance(data, str):
        return data
    return data.strip()


def uppercase(data: None | int | str) -> None | int | str:
    if not isinstance(data, str):
        return data
    return data.upper()


def lowercase(data: None | int | str) -> None | int | str:
    if not isinstance(data, str):
        return data
    return data.lower()



""""""
from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable, Dict

if TYPE_CHECKING:
    from .extension import Admin


class AdminPanel:
    """Base classe for admin panels.

    Currently this class does nothing. It may be useful in the future
    either as just a marker interface (for automatic plugin discovery /
    registration), or to add some common functionnalities. Otherwise, it
    will be removed.
    """

    id: str = ""
    label: str = ""
    icon: str = ""
    admin: Admin

    def url_value_preprocess(self, endpoint: str, view_args: dict[Any, Any]):
        """Panel can preprocess values for their views.

        This method is called only if the endpoint is for `get()`, `post()`, or
        one of the views installed with `install_additional_rules`.

        This is also the right place to add items to the breadcrumbs.
        """

    def install_additional_rules(self, add_url_rule: Callable):
        """This method can be redefined in subclasses to install custom url
        rules.

        All rules are relative to panel 'base' rule, don't prefix rules with panel
        id, it will be done by `add_url_rule`.

        :param add_url_rule: function to use to add url rules, same interface as
            :meth:`flask.blueprint.Blueprint.add_url_rule`.
        """



from __future__ import annotations

from abilian.core.util import Pagination, slugify


def test_pagination_1():
    p = Pagination(1, 10, 10)
    pages = list(p.iter_pages())
    assert pages == [1]
    assert not p.has_prev
    assert not p.has_next
    assert p.prev is None
    assert p.next is None


def test_pagination_2():
    p = Pagination(1, 10, 20)
    pages = list(p.iter_pages())
    assert pages == [1, 2]
    assert not p.has_prev
    assert p.has_next
    assert p.prev is None
    assert p.next == 2


def test_pagination_3():
    p = Pagination(1, 10, 100)
    pages = list(p.iter_pages())
    assert pages == [1, 2, 3, 4, 5, None, 9, 10]
    assert not p.has_prev
    assert p.has_next
    assert p.prev is None
    assert p.next == 2


def test_pagination_4():
    p = Pagination(10, 10, 100)
    pages = list(p.iter_pages())
    assert pages == [1, 2, None, 8, 9, 10]
    assert p.has_prev
    assert not p.has_next
    assert p.prev == 9
    assert p.next is None


def test_slugify_basic():
    slug = slugify("a b c")
    assert slug == "a-b-c"
    assert isinstance(slug, str)
    assert slugify(slug) == "a-b-c"  # idempotent


def test_slugify_separator():
    slug = slugify("a-b++ c-+", "+")
    assert slug == "a+b+c"


def test_slugify_non_ascii():
    slug = slugify("C'est l'été !")
    assert slug == "c-est-l-ete"

    # with a special space character
    slug = slugify("a_b\u205fc")  # U+205F: MEDIUM MATHEMATICAL SPACE
    assert slug == "a-b-c"

    # with non-ascii translatable chars, like EN DASH U+2013 (–) and EM DASH
    # U+2014 (—).
    # this test fails if regexp subst is done after Unicode normalization
    assert slugify("a\u2013b\u2014c") == "a-b-c"



from __future__ import annotations

import runpy

import click
from flask.cli import with_appcontext

from abilian.core.extensions import db
from abilian.core.models.subjects import User
from abilian.services import get_service


@click.command()
@with_appcontext
def initdb():
    db.create_all()


@click.command()
@with_appcontext
def dropdb():
    """Drop the application DB."""
    confirm = input("Are you sure you want to drop the database? (Y/N) ")
    print(f"Dropping DB using engine: {db}")
    if confirm.lower() == "y":
        # with current_app.app_context():
        db.drop_all()


@click.command()
@click.argument("path", type=click.Path(exists=True))
@with_appcontext
def script(path):
    """Run given script in the app context."""
    runpy.run_path(path, run_name="__main__")


@click.command()
@click.argument("email")
@click.argument("password")
@click.option("--role")
@click.option("--name")
@click.option("--first_name")
@with_appcontext
def createuser(email, password, role=None, name=None, first_name=None):
    """Create new user."""

    if User.query.filter(User.email == email).count() > 0:
        print(f"A user with email '{email}' already exists, aborting.")
        return

    # if password is None:
    #     password = prompt_pass("Password")

    user = User(
        email=email,
        password=password,
        last_name=name,
        first_name=first_name,
        can_login=True,
    )
    db.session.add(user)

    if role in ("admin",):
        # FIXME: add other valid roles
        security = get_service("security")
        security.grant_role(user, role)

    db.session.commit()
    print(f"User {email} added")


# def createadmin(email, password, name=None, first_name=None):
#     """Create new administrator.
#
#     Same as `createuser --role='admin'`.
#     """
#     createuser(email, password, role="admin", name=name, first_name=first_name)


#
# # coding=utf-8
# import logging
# import runpy
# from pprint import pformat
#
# import click
# import sqlalchemy as sa
# import sqlalchemy.exc
# from flask import current_app
# from flask.cli import AppGroup, FlaskGroup, with_appcontext
# from flask_script import Manager, prompt_pass
# from six import text_type
# from six.moves import input, urllib
#
# from abilian.core.extensions import db
# from abilian.core.logging import patch_logger
# from abilian.core.models.subjects import User
# from abilian.services import get_service
# from abilian.services.security import Role
#
# __all__ = ["manager", "logger"]
#
# # Setup basic logging capabilities in case logging is not yet set up. From doc:
# # This function does nothing if the root logger already has handlers configured
# # for it.
# # Allow "print" statements to be replaced by a logging statements
# logging.basicConfig()
# logger = logging.getLogger("")
#
# # PATCH flask_script.Manager.run to force creation of app before run() is
# # called. In default implementation, the arg parser is created before the Flask
# # application. So we can't use app.script_manager to add commands from
# # plugins. If app is created before the arg parser, plugin commands are properly
# # registered
# _flask_script_manager_run = Manager.run
#
#
# def _manager_run(self, *args, **kwargs):
#     self()
#     if "sentry" in self.app.extensions:
#         client = self.app.extensions["sentry"].client
#         client.tags["process_type"] = "shell"
#
#     return _flask_script_manager_run(self, *args, **kwargs)
#
#
# patch_logger.info(Manager.run)
# Manager.run = _manager_run
#
# #: ``flask_script.Manager`` instance for abilian commands
# manager = Manager(usage="Abilian base commands")
#
#
# def _log_config(config):
#     lines = ["Application configuration:"]
#
#     if config.get("CONFIGURED"):
#         settings = get_service("settings")
#         try:
#             db_settings = set(settings.namespace("config").keys())
#         except sa.exc.ProgrammingError:
#             # there is config.py, db uri, but maybe "initdb" has yet to be run
#             db_settings = {}
#     else:
#         db_settings = {}
#
#     for k, v in sorted(config.items()):
#         prefix = "    "
#         if k in db_settings:
#             prefix = "  * "
#         indent = len(k) + 3
#         width = 80 - indent
#         v = pformat(v, width=width).replace("\n", "\n" + " " * indent)
#         lines.append(f"{prefix}{k}: {v}")
#     logger.info("\n".join(lines))
#
#
# def log_config(config):
#     original_level = logger.level
#     logger.setLevel(logging.INFO)
#     try:
#         return _log_config(config)
#     finally:
#         logger.setLevel(original_level)
#
#
# @manager.option("-p", "--port", dest="port", help="listening port", default=5000)
# @manager.option(
#     "--show-config",
#     dest="show_config",
#     action="store_const",
#     const=True,
#     default=False,
#     help="show application configuration on startup",
# )
# @manager.option(
#     "--ssl",
#     dest="ssl",
#     action="store_const",
#     default=False,
#     const=True,
#     help="Enable werkzeug SSL",
# )
# def run(port, show_config, ssl):
#     """Like runserver.
#
#     May also print application configuration if used with
#     --show-config.
#     """
#     app = current_app
#     options = {}
#     if show_config:
#         log_config(app.config)
#
#     # TODO: pass host and debug as params to
#     host = "0.0.0.0"
#     debug = app.config.get("DEBUG")
#     port = int(port or app.config.get("PORT", 5000))
#
#     if ssl:
#         options["ssl_context"] = "adhoc"
#
#     app.run(host=host, debug=debug, port=port, **options)
#
#
# @manager.command
# def routes():
#     """Show all the routes registered in Flask."""
#     output = []
#     for rule in current_app.url_map.iter_rules():
#         methods = ",".join(rule.methods)
#         path = urllib.parse.unquote(rule.rule)
#         output.append((rule.endpoint, methods, path))
#
#     for endpoint, methods, path in sorted(output):
#         print(f"{endpoint:40s} {methods:25s} {path}")
#
#
# @email_opt
# @password_opt
# def passwd(email, password=None):
#     """Change the password for the given user."""
#     user = User.query.filter(User.email == email).one()
#     if password is None:
#         password = prompt_pass("New password: ")
#
#     user.set_password(password)
#     db.session.commit()
#     print(f"Password updated for user {email}")
#



""""""
from __future__ import annotations

from typing import Iterator, cast

import sqlalchemy as sa
from pytest import fixture
from sqlalchemy.orm import Session

from abilian.app import Application
from abilian.core.entities import Entity
from abilian.services import get_service
from abilian.services.indexing.service import WhooshIndexService


class IndexedContact(Entity):
    # default is 'test_service.IndexedContact'
    entity_type = "abilian.services.indexing.IndexedContact"
    name = sa.Column(sa.UnicodeText)


@fixture
def svc(app: Application) -> Iterator[WhooshIndexService]:
    _svc = cast(WhooshIndexService, get_service("indexing"))
    with app.app_context():
        _svc.start()
        yield _svc


def test_app_state(app: Application, svc: WhooshIndexService):
    state = svc.app_state
    assert IndexedContact in state.indexed_classes
    assert IndexedContact.entity_type in svc.adapted
    assert IndexedContact.entity_type in state.indexed_fqcn


def test_index_only_after_final_commit(
    app: Application, session: Session, svc: WhooshIndexService
):
    contact = IndexedContact(name="John Doe")

    state = svc.app_state

    session.begin(nested=True)
    assert state.to_update == []

    session.add(contact)
    # no commit: model is in wait queue
    session.flush()
    assert state.to_update == [("new", contact)]

    # commit but in a sub transaction: model still in wait queue
    session.commit()
    assert state.to_update == [("new", contact)]

    # 'final' commit: models sent for indexing update
    session.commit()
    assert state.to_update == []


def test_clear(app: Application, svc: WhooshIndexService):
    # just check no exception happens
    svc.clear()

    # check no double stop (would raise AssertionError from service base)
    svc.start()
    svc.stop()
    svc.clear()



"""Base Flask application class, used by tests or to be extended in real
applications."""
from __future__ import annotations

import errno
import importlib
import logging.config
import os
import sys
import warnings
from functools import partial
from itertools import chain, count
from pathlib import Path
from typing import Any, Callable, Collection, Dict, Optional

import jinja2
import sqlalchemy as sa
import sqlalchemy.exc
from flask import (
    Blueprint,
    Flask,
    abort,
    appcontext_pushed,
    g,
    request,
    request_started,
)
from flask.config import Config, ConfigAttribute
from flask.helpers import locked_cached_property
from flask_migrate import Migrate
from flask_talisman import DEFAULT_CSP_POLICY, Talisman
from sqlalchemy.orm.attributes import NEVER_SET

import abilian.core.util
import abilian.i18n
from abilian.config import default_config
from abilian.core import extensions, signals
from abilian.core.celery import FlaskCelery
from abilian.services import (
    Service,
    activity_service,
    antivirus,
    audit_service,
    auth_service,
    conversion_service,
    index_service,
    preferences_service,
    repository_service,
    security_service,
    session_repository_service,
    settings_service,
    vocabularies_service,
)
from abilian.services.security import Anonymous
from abilian.services.security.models import Role
from abilian.web import csrf
from abilian.web.action import actions
from abilian.web.admin import Admin
from abilian.web.assets import AssetManagerMixin
from abilian.web.blueprints import allow_access_for_roles
from abilian.web.errors import ErrorManagerMixin
from abilian.web.hooks import init_hooks
from abilian.web.jinja import JinjaManagerMixin
from abilian.web.nav import BreadcrumbItem
from abilian.web.util import send_file_from_directory
from abilian.web.views import Registry as ViewRegistry

logger = logging.getLogger(__name__)
db = extensions.db
__all__ = ["create_app", "Application", "ServiceManager"]

# Silence those warnings for now.
warnings.simplefilter("ignore", category=sa.exc.SAWarning)


class ServiceManager:
    """Mixin that provides lifecycle (register/start/stop) support for
    services."""

    services: dict[str, Service]

    def __init__(self):
        self.services = {}

    def start_services(self):
        for svc in self.services.values():
            svc.start()

    def stop_services(self):
        for svc in self.services.values():
            svc.stop()


class PluginManager:
    """Mixin that provides support for loading plugins."""

    config: Config

    #: Custom apps may want to always load some plugins: list them here.
    APP_PLUGINS = (
        "abilian.web.search",
        "abilian.web.tags",
        "abilian.web.comments",
        "abilian.web.uploads",
        "abilian.web.attachments",
    )

    def register_plugin(self, name: str):
        """Load and register a plugin given its package name."""
        logger.info(f"Registering plugin: {name}")
        module = importlib.import_module(name)
        module.register_plugin(self)  # type: ignore

    def register_plugins(self):
        """Load plugins listed in config variable 'PLUGINS'."""
        registered = set()
        for plugin_fqdn in chain(self.APP_PLUGINS, self.config["PLUGINS"]):
            if plugin_fqdn not in registered:
                self.register_plugin(plugin_fqdn)
                registered.add(plugin_fqdn)


class Application(
    ServiceManager,
    PluginManager,
    AssetManagerMixin,
    ErrorManagerMixin,
    JinjaManagerMixin,
    Flask,
):
    """Base application class.

    Extend it in your own app.
    """

    default_config = default_config

    #: If True all views will require by default an authenticated user, unless
    #: Anonymous role is authorized. Static assets are always public.
    private_site = ConfigAttribute("PRIVATE_SITE")

    #: instance of :class:`.web.views.registry.Registry`.
    default_view: ViewRegistry

    #: json serializable dict to land in Javascript under Abilian.api
    js_api: dict[str, Any]

    #: celery app class
    celery_app_cls = FlaskCelery

    def __init__(self, name: Any | None = None, *args: Any, **kwargs: Any):
        name = name or __name__

        Flask.__init__(self, name, *args, **kwargs)

        ServiceManager.__init__(self)
        PluginManager.__init__(self)
        JinjaManagerMixin.__init__(self)

        self.default_view = ViewRegistry()
        self.js_api = {}

    def setup(self, config: type | None):
        self.configure(config)

        # At this point we have loaded all external config files:
        # SQLALCHEMY_DATABASE_URI is definitively fixed (it cannot be defined in
        # database AFAICT), and LOGGING_FILE cannot be set in DB settings.
        self.setup_logging()

        appcontext_pushed.connect(self.install_id_generator)

        if not self.testing:
            self.init_sentry()

        # time to load config bits from database: 'settings'
        # First init required stuff: db to make queries, and settings service
        extensions.db.init_app(self)
        settings_service.init_app(self)

        self.register_jinja_loaders(jinja2.PackageLoader("abilian.web"))
        self.init_assets()
        self.install_default_handlers()

        with self.app_context():
            self.init_extensions()
            self.register_plugins()
            self.add_access_controller(
                "static", allow_access_for_roles(Anonymous), endpoint=True
            )
            # debugtoolbar: this is needed to have it when not authenticated
            # on a private site. We cannot do this in init_debug_toolbar,
            # since auth service is not yet installed.
            self.add_access_controller(
                "debugtoolbar", allow_access_for_roles(Anonymous)
            )
            self.add_access_controller(
                "_debug_toolbar.static",
                allow_access_for_roles(Anonymous),
                endpoint=True,
            )

        # TODO: maybe reenable later
        # self.maybe_register_setup_wizard()

        self._finalize_assets_setup()

        # At this point all models should have been imported: time to configure
        # mappers. Normally Sqlalchemy does it when needed but mappers may be
        # configured inside sa.orm.class_mapper() which hides a
        # misconfiguration: if a mapper is misconfigured its exception is
        # swallowed by class_mapper(model) results in this laconic
        # (and misleading) message: "model is not mapped"
        sa.orm.configure_mappers()

        signals.components_registered.send(self)

        request_started.connect(self.setup_nav_and_breadcrumbs)
        init_hooks(self)

        # Initialize Abilian core services.
        # Must come after all entity classes have been declared.
        # Inherited from ServiceManager. Will need some configuration love
        # later.
        if not self.testing:
            with self.app_context():
                self.start_services()

        setup(self)

    def setup_nav_and_breadcrumbs(self, app: Flask):
        """Listener for `request_started` event.

        If you want to customize first items of breadcrumbs, override
        :meth:`init_breadcrumbs`
        """
        g.nav = {"active": None}  # active section
        g.breadcrumb = []
        self.init_breadcrumbs()

    def init_breadcrumbs(self):
        """Insert the first element in breadcrumbs.

        This happens during `request_started` event, which is triggered
        before any url_value_preprocessor and `before_request` handlers.
        """
        g.breadcrumb.append(BreadcrumbItem(icon="home", url=f"/{request.script_root}"))

    # TODO: remove
    def install_id_generator(self, sender: Flask, **kwargs: Any):
        g.id_generator = count(start=1)

    def configure(self, config: type | None):
        if config:
            self.config.from_object(config)

        # Setup babel config
        languages = self.config["BABEL_ACCEPT_LANGUAGES"]
        languages = tuple(
            lang for lang in languages if lang in abilian.i18n.VALID_LANGUAGES_CODE
        )
        self.config["BABEL_ACCEPT_LANGUAGES"] = languages

        # This needs to be done dynamically
        if not self.config.get("SESSION_COOKIE_NAME"):
            self.config["SESSION_COOKIE_NAME"] = f"{self.name}-session"

        if not self.config.get("FAVICO_URL"):
            self.config["FAVICO_URL"] = self.config.get("LOGO_URL")

        if not self.debug and self.config["SECRET_KEY"] == "CHANGEME":
            logger.error("You must change the default secret config ('SECRET_KEY')")
            sys.exit()

    def check_instance_folder(self, create=False):
        """Verify instance folder exists, is a directory, and has necessary
        permissions.

        :param:create: if `True`, creates directory hierarchy

        :raises: OSError with relevant errno if something is wrong.
        """
        path = Path(self.instance_path)
        err = None
        eno = 0

        if not path.exists():
            if create:
                logger.info("Create instance folder: %s", path)
                path.mkdir(0o775, parents=True)
            else:
                err = "Instance folder does not exists"
                eno = errno.ENOENT
        elif not path.is_dir():
            err = "Instance folder is not a directory"
            eno = errno.ENOTDIR
        elif not os.access(str(path), os.R_OK | os.W_OK | os.X_OK):
            err = 'Require "rwx" access rights, please verify permissions'
            eno = errno.EPERM

        if err:
            raise OSError(eno, err, str(path))

    @locked_cached_property
    def data_dir(self) -> Path:
        path = Path(self.instance_path, "data")
        if not path.exists():
            path.mkdir(0o775, parents=True)

        return path

    def init_extensions(self):
        """Initialize flask extensions, helpers and services."""
        extensions.redis.init_app(self)
        extensions.mail.init_app(self)
        extensions.deferred_js.init_app(self)
        extensions.upstream_info.extension.init_app(self)
        actions.init_app(self)

        # auth_service installs a `before_request` handler (actually it's
        # flask-login). We want to authenticate user ASAP, so that sentry and
        # logs can report which user encountered any error happening later,
        # in particular in a before_request handler (like csrf validator)
        auth_service.init_app(self)

        # webassets
        self.setup_asset_extension()
        self.register_base_assets()

        # Babel (for i18n)
        babel = abilian.i18n.babel
        # Temporary (?) workaround
        babel.locale_selector_func = None
        babel.timezone_selector_func = None

        babel.init_app(self)
        babel.add_translations("wtforms", translations_dir="locale", domain="wtforms")
        babel.add_translations("abilian")
        babel.localeselector(abilian.i18n.localeselector)
        babel.timezoneselector(abilian.i18n.timezoneselector)

        # Flask-Migrate
        Migrate(self, db)

        # CSRF by default
        if self.config.get("WTF_CSRF_ENABLED"):
            extensions.csrf.init_app(self)
            self.extensions["csrf"] = extensions.csrf
            extensions.abilian_csrf.init_app(self)

        self.register_blueprint(csrf.blueprint)

        # images blueprint
        from .web.views.images import blueprint as images_bp

        self.register_blueprint(images_bp)

        # Abilian Core services
        security_service.init_app(self)
        repository_service.init_app(self)
        session_repository_service.init_app(self)
        audit_service.init_app(self)
        index_service.init_app(self)
        activity_service.init_app(self)
        preferences_service.init_app(self)
        conversion_service.init_app(self)
        vocabularies_service.init_app(self)
        antivirus.init_app(self)

        from .web.preferences.user import UserPreferencesPanel

        preferences_service.register_panel(UserPreferencesPanel(), self)

        from .web.coreviews import users

        self.register_blueprint(users.blueprint)

        # Admin interface
        Admin().init_app(self)

        # Celery async service
        # this allows all shared tasks to use this celery app
        if getattr(self, "celery_app_cls", None):
            celery_app = self.extensions["celery"] = self.celery_app_cls()
            # force reading celery conf now - default celery app will
            # also update our config with default settings
            celery_app.conf  # noqa
            celery_app.set_default()

        # dev helper
        if self.debug:
            # during dev, one can go to /http_error/403 to see rendering of 403
            http_error_pages = Blueprint("http_error_pages", __name__)

            @http_error_pages.route("/<int:code>")
            def error_page(code):
                """Helper for development to show 403, 404, 500..."""
                abort(code)

            self.register_blueprint(http_error_pages, url_prefix="/http_error")

    def add_url_rule_with_role(
        self,
        rule: str,
        endpoint: str,
        view_func: Callable,
        roles: Collection[Role] = (),
        **options: Any,
    ):
        """See :meth:`Flask.add_url_rule`.

        If `roles` parameter is present, it must be a
        :class:`abilian.service.security.models.Role` instance, or a list of
        Role instances.
        """
        self.add_url_rule(rule, endpoint, view_func, **options)

        if roles:
            self.add_access_controller(
                endpoint, allow_access_for_roles(roles), endpoint=True
            )

    def add_access_controller(self, name: str, func: Callable, endpoint: bool = False):
        """Add an access controller.

        If `name` is None it is added at application level, else if is
        considered as a blueprint name. If `endpoint` is True then it is
        considered as an endpoint.
        """
        auth_state = self.extensions[auth_service.name]

        if endpoint:
            if not isinstance(name, str):
                msg = f"{repr(name)} is not a valid endpoint name"
                raise ValueError(msg)

            auth_state.add_endpoint_access_controller(name, func)
        else:
            auth_state.add_bp_access_controller(name, func)

    def add_static_url(
        self, url_path: str, directory: str, endpoint: str, roles: Collection[Role] = ()
    ):
        """Add a new url rule for static files.

        :param url_path: subpath from application static url path. No heading
            or trailing slash.
        :param directory: directory to serve content from.
        :param endpoint: flask endpoint name for this url rule.

        Example::

           app.add_static_url('myplugin',
                              '/path/to/myplugin/resources',
                              endpoint='myplugin_static')

        With default setup it will serve content from directory
        `/path/to/myplugin/resources` from url `http://.../static/myplugin`
        """
        url_path = f"{self.static_url_path}/{url_path}/<path:filename>"
        self.add_url_rule_with_role(
            url_path,
            endpoint=endpoint,
            view_func=partial(send_file_from_directory, directory=directory),
            roles=roles,
        )
        self.add_access_controller(
            endpoint, allow_access_for_roles(Anonymous), endpoint=True
        )


def setup(app: Flask):
    config = app.config

    # CSP
    if not app.debug:
        csp = config.get("CONTENT_SECURITY_POLICY", DEFAULT_CSP_POLICY)
        Talisman(app, content_security_policy=csp)

    # Debug Toolbar
    init_debug_toolbar(app)


def init_debug_toolbar(app: Flask):
    if not app.debug or app.testing:
        return

    try:
        from flask_debugtoolbar import DebugToolbarExtension
    except ImportError:
        logger.warning("Running in debug mode but flask_debugtoolbar is not installed.")
        return

    dbt = DebugToolbarExtension()
    default_config = dbt._default_config(app)
    init_dbt = dbt.init_app

    if "DEBUG_TB_PANELS" not in app.config:
        # add our panels to default ones
        app.config["DEBUG_TB_PANELS"] = list(default_config["DEBUG_TB_PANELS"])
    init_dbt(app)
    for view_name in app.view_functions:
        if view_name.startswith("debugtoolbar."):
            extensions.csrf.exempt(app.view_functions[view_name])


def create_app(
    config: type | None = None, app_class: type = Application, **kw: Any
) -> Application:
    app = app_class(**kw)
    app.setup(config=config)

    # This is currently called from app.setup()
    # setup(app)

    return app



"""Social graph API.

May or may not be needed, since the followers / followees notions are
already implemented in the core model.
"""


from __future__ import annotations


class FollowService:
    def get_followers(self, user):
        pass

    def get_followees(self, user):
        pass



""""""
from __future__ import annotations

from datetime import datetime
from typing import Any, Dict

import sqlalchemy as sa
from whoosh.fields import NUMERIC, TEXT, Schema

from abilian.core.entities import Entity
from abilian.core.models.base import SEARCHABLE, IdMixin
from abilian.core.models.base import Indexable as CoreIndexable
from abilian.core.models.base import Model
from abilian.services.indexing.adapter import SAAdapter


class SANotAdaptable:
    pass


class SANotIndexable(IdMixin, Model):
    __tablename__ = "sa_not_indexable"
    __indexable__ = False


class Indexable(IdMixin, CoreIndexable, Model):
    __tablename__ = "sa_indexable"
    __index_to__ = (("related.name", ("name", "text")), ("related.description", "text"))

    num = sa.Column(sa.Integer, info=SEARCHABLE | {"index_to": (("num", NUMERIC()),)})


class SubclassEntityIndexable(Entity):
    pass


def test_can_adapt():
    assert not SAAdapter.can_adapt(SANotAdaptable)
    assert SAAdapter.can_adapt(SANotIndexable)
    assert SAAdapter.can_adapt(Indexable)
    assert SAAdapter.can_adapt(Entity)


def test_build_attrs_1():
    schema = Schema()
    adapter = SAAdapter(SANotIndexable, schema)
    assert not adapter.indexable
    assert adapter.doc_attrs == {}


def test_build_attrs_2():
    schema = Schema()
    adapter = SAAdapter(Entity, schema)
    assert adapter.indexable == False


def test_build_attrs_3():
    schema = Schema()
    adapter = SAAdapter(SubclassEntityIndexable, schema)
    assert adapter.indexable
    assert set(adapter.doc_attrs) == {
        "allowed_roles_and_users",
        "created_at",
        "creator",
        "creator_name",
        "id",
        "name",
        "name_prefix",
        "object_key",
        "object_type",
        "owner",
        "owner_name",
        "slug",
        "tag_ids",
        "tag_text",
        "text",
        "updated_at",
    }
    assert all(lambda f: callable(f) for f in adapter.doc_attrs.values())

    assert set(schema.names()) == {
        "allowed_roles_and_users",
        "created_at",
        "creator",
        "creator_name",
        "id",
        "name",
        "name_prefix",
        "object_key",
        "object_type",
        "owner",
        "owner_name",
        "slug",
        "tag_ids",
        "tag_text",
        "text",
        "updated_at",
    }


def test_build_attrs_4():
    schema = Schema(id=NUMERIC(bits=64, signed=False, stored=True, unique=True))
    adapter = SAAdapter(Indexable, schema)
    assert adapter.indexable
    assert set(adapter.doc_attrs) == {
        "id",
        "text",
        "num",
        "name",
        "object_type",
        "object_key",
    }
    assert all(lambda f: callable(f) for f in adapter.doc_attrs.values())

    assert set(schema.names()) == {
        "id",
        "text",
        "num",
        "name",
        "object_type",
        "object_key",
    }
    assert isinstance(schema["text"], TEXT)
    assert isinstance(schema["num"], NUMERIC)


def test_get_document(app, db):
    schema = Schema()
    adapter = SAAdapter(SubclassEntityIndexable, schema)
    expected = {
        "id": 2,
        "name": "entity name",
        "created_at": datetime(2013, 11, 28, 16, 17, 0),
        "updated_at": datetime(2013, 11, 29, 12, 17, 58),
    }
    obj = SubclassEntityIndexable(**expected)
    obj.slug = "entity-name"

    expected["object_type"] = "test_adapter.SubclassEntityIndexable"
    expected["object_key"] = "test_adapter.SubclassEntityIndexable:2"
    expected["text"] = "entity name"
    expected["slug"] = "entity-name"
    expected["name_prefix"] = "entity name"
    expected["allowed_roles_and_users"] = "role:admin"

    with app.test_request_context():
        assert adapter.get_document(obj) == expected


def test_get_document_with_schema():
    # test retrieve related attributes
    schema = Schema(id=NUMERIC(bits=64, signed=False, stored=True, unique=True))
    adapter = SAAdapter(Indexable, schema)
    expected: dict[str, Any] = {"id": 1, "num": 42}
    obj = Indexable(**expected)
    obj.related = type("Related", (object,), {"name": None})()
    expected["name"] = obj.related.name = "related name"
    obj.related.description = "description text"
    expected["text"] = f"{obj.related.name} {obj.related.description}"
    doc = adapter.get_document(obj)

    assert set(doc) == {"id", "name", "num", "text", "object_type", "object_key"}
    assert doc["id"] == 1
    assert doc["num"] == 42
    assert doc["name"] == "related name"
    assert "related name" in doc["text"]
    assert "description text" in doc["text"]



""""""
from __future__ import annotations

import sqlalchemy as sa
from wtforms.fields import BooleanField, StringField
from wtforms.validators import ValidationError

from abilian.core.models.subjects import Group
from abilian.i18n import _, _l
from abilian.services.security.models import Role
from abilian.web.forms import Form, widgets
from abilian.web.forms.fields import QuerySelect2Field, Select2MultipleField
from abilian.web.forms.filters import strip
from abilian.web.forms.validators import optional, required


class BaseUserAdminForm(Form):

    email = StringField(
        _l("Email"),
        description=_l("Users log in with their email address."),
        view_widget=widgets.EmailWidget(),
        filters=(strip,),
        validators=[required()],
    )
    first_name = StringField(
        _l("First Name"),
        description=_l("ex: John"),
        filters=(strip,),
        validators=[required()],
    )
    last_name = StringField(
        _l("Last Name"),
        description=_l("ex: Smith"),
        filters=(strip,),
        validators=[required()],
    )

    can_login = BooleanField(
        _l("Login enabled"),
        description=_l("If unchecked, user will not be able to connect."),
        widget=widgets.BooleanWidget(),
    )

    groups = QuerySelect2Field(
        _l("Groups"),
        validators=(optional(),),
        multiple=True,
        collection_class=set,
        query_factory=lambda: Group.query.order_by(sa.sql.func.lower(Group.name).asc()),
        get_label="name",
    )

    roles = Select2MultipleField(
        _l("Roles"),
        description=_l(
            "Prefer groups to manage access rights. Directly assigning roles "
            "to users is possible but discouraged."
        ),
        choices=lambda: [(r.name, r.label) for r in Role.assignable_roles()],
    )

    password = StringField(
        _l("New Password"),
        description=_l("If empty the current password will not be changed."),
        widget=widgets.PasswordInput(autocomplete="off"),
    )


class UserAdminForm(BaseUserAdminForm):

    confirm_password = StringField(
        _l("Confirm new password"), widget=widgets.PasswordInput(autocomplete="off")
    )

    def validate_password(self, field):
        pwd = field.data
        confirmed = self["confirm_password"].data

        if pwd != confirmed:
            raise ValidationError(
                _(
                    "Passwords differ. Ensure you have typed same password in both"
                    ' "password" field and "confirm password" field.'
                )
            )


class UserCreateForm(BaseUserAdminForm):

    password = StringField(
        _l("Password"),
        description=_l("If empty a random password will be generated."),
        widget=widgets.PasswordInput(autocomplete="off"),
    )



""""""
from __future__ import annotations

import abc

import sqlalchemy as sa
import sqlalchemy.event
from sqlalchemy import Column, ForeignKey, Integer, UnicodeText
from sqlalchemy.orm import backref, relationship

from abilian.core.entities import Entity, EntityQuery

from .blob import Blob

#: name of backref on target :class:`Entity` object
ATTRIBUTE = "__attachments__"


class SupportAttachment(metaclass=abc.ABCMeta):
    pass


def register(cls):
    """Register an :class:`~.Entity` as an attachable class.

    Can be used as a class decorator:

    .. code-block:: python

      @attachment.register
      class MyContent(Entity):
          ....
    """
    if not issubclass(cls, Entity):
        raise ValueError("Class must be a subclass of abilian.core.entities.Entity")

    SupportAttachment.register(cls)
    return cls


def supports_attachments(obj):
    """
    :param obj: a class or instance

    :returns: True is obj supports attachments.
    """
    if isinstance(obj, type):
        return issubclass(obj, SupportAttachment)

    if not isinstance(obj, SupportAttachment):
        return False

    if obj.id is None:
        return False

    return True


def for_entity(obj, check_support_attachments=False):
    """Return attachments on an entity."""
    if check_support_attachments and not supports_attachments(obj):
        return []

    return getattr(obj, ATTRIBUTE)


class AttachmentQuery(EntityQuery):
    def all(self):
        return EntityQuery.all(self.order_by(Attachment.created_at))


class Attachment(Entity):
    """An Attachment owned by an :class:`Entity`."""

    __auditable_entity__ = ("entity", "attachment", ("id", "name"))

    entity_id = Column(Integer, ForeignKey(Entity.id), nullable=False)

    #: owning entity
    entity = relationship(
        Entity,
        lazy="immediate",
        foreign_keys=[entity_id],
        backref=backref(
            ATTRIBUTE,
            lazy="select",
            order_by="Attachment.created_at",
            cascade="all, delete-orphan",
        ),
    )

    blob_id = Column(Integer, sa.ForeignKey(Blob.id), nullable=False)
    #: file. Stored in a :class:`Blob`
    blob = relationship(Blob, cascade="all, delete", foreign_keys=[blob_id])

    description = Column(UnicodeText(), nullable=False, default="", server_default="")

    query_class = AttachmentQuery

    def __repr__(self):
        class_ = self.__class__
        mod_ = class_.__module__
        classname = class_.__name__
        return "<{}.{} instance at 0x{:x} entity id={!r}>".format(
            mod_, classname, id(self), self.entity_id
        )


@sa.event.listens_for(Attachment, "before_insert", propagate=True)
@sa.event.listens_for(Attachment, "before_update", propagate=True)
def set_attachment_name(mapper, connection, target):
    if target.name:
        return

    blob = target.blob
    if not blob:
        return

    filename = blob.meta.get("filename")
    if filename is not None:
        target.name = filename



""""""
from __future__ import annotations

import typing
from typing import Dict

from flask import current_app, send_file
from flask_login import current_user
from flask_wtf.file import FileField, file_required
from werkzeug.exceptions import BadRequest, NotFound
from werkzeug.utils import secure_filename

from abilian.core.models.subjects import User
from abilian.core.util import unwrap
from abilian.web import csrf, url_for
from abilian.web.blueprints import Blueprint
from abilian.web.forms import Form
from abilian.web.views import JSONView, View

if typing.TYPE_CHECKING:
    from abilian.web.uploads import FileUploadsExtension


bp = Blueprint("uploads", __name__, url_prefix="/upload")


class UploadForm(Form):

    file = FileField(validators=(file_required(),))


class BaseUploadsView:
    uploads: FileUploadsExtension
    user: User

    def prepare_args(self, args, kwargs):
        args, kwargs = super().prepare_args(args, kwargs)
        self.uploads = current_app.extensions["uploads"]
        self.user = unwrap(current_user)
        return args, kwargs


class NewUploadView(BaseUploadsView, JSONView):
    """Upload a new file."""

    methods = ["POST", "PUT"]
    decorators = (csrf.support_graceful_failure,)

    #: file handle to be returned
    handle = None

    def data(self, *args, **kwargs) -> dict:
        return {"handle": self.handle, "url": url_for(".handle", handle=self.handle)}

    def post(self, *args, **kwargs):
        form = UploadForm()

        if not form.validate():
            raise BadRequest("File is missing.")

        uploaded = form["file"].data
        filename = secure_filename(uploaded.filename)
        mimetype = uploaded.mimetype
        self.handle = self.uploads.add_file(
            self.user, uploaded, filename=filename, mimetype=mimetype
        )
        return self.get(*args, **kwargs)

    def put(self, *args, **kwargs):
        return self.post(*args, **kwargs)


bp.add_url_rule("/", view_func=NewUploadView.as_view("new_file"))


class UploadView(BaseUploadsView, View):
    """Manage an uploaded file: download, delete."""

    methods = ["GET", "DELETE"]
    decorators = (csrf.support_graceful_failure,)

    def get(self, handle, *args, **kwargs):
        file_obj = self.uploads.get_file(self.user, handle)

        if file_obj is None:
            raise NotFound()

        metadata = self.uploads.get_metadata(self.user, handle)
        filename = metadata.get("filename", handle)
        content_type = metadata.get("mimetype")
        stream = file_obj.open("rb")

        return send_file(
            stream,
            as_attachment=True,
            attachment_filename=filename,
            mimetype=content_type,
            cache_timeout=0,
            add_etags=False,
        )

    def delete(self, handle, *args, **kwargs) -> dict:
        if self.uploads.get_file(self.user, handle) is None:
            raise NotFound()
        self.uploads.remove_file(self.user, handle)
        return {"success": True}


bp.add_url_rule("/<string:handle>", view_func=UploadView.as_view("handle"))



from __future__ import annotations


class PrivateMessage:
    # TODO
    pass


class PrivateMessageService:
    def messages(self, user=None, page=0, per_page=50, filter=None):
        """Return the list of messages received by the given user (or current
        user if user=None)."""

    def send(self, dest_user, subject, body, url=None):
        """Send a message to the destination user."""


class MessageService:
    # TODO
    pass



from __future__ import annotations

from typing import TYPE_CHECKING, Any, List, Optional

import sqlalchemy as sa
from sqlalchemy.orm import object_session

from abilian.core.entities import Entity
from abilian.core.models.subjects import User
from abilian.core.signals import activity
from abilian.services import Service

from .models import ActivityEntry

if TYPE_CHECKING:
    pass

__all__ = ["ActivityService"]


class ActivityService(Service):
    name = "activity"

    def start(self, ignore_state: bool = False):
        super().start(ignore_state)
        activity.connect(self.log_activity)

    def stop(self, ignore_state: bool = False):
        super().stop(ignore_state)
        activity.disconnect(self.log_activity)

    def log_activity(
        self,
        sender: None,
        actor: User,
        verb: str,
        object: Any,
        target: Entity | None = None,
    ):
        assert self.running
        if not isinstance(object, Entity):
            # generic forms may send signals inconditionnaly. For now we have activity
            # only for Entities
            return

        session = object_session(object)
        kwargs = {"actor": actor, "verb": verb, "object_type": object.entity_type}

        if sa.inspect(object).deleted:
            # object is in deleted state: flush has occurred, don't reference it or
            # we'll have an error when adding entry to session
            kwargs["object_id"] = object.id
        else:
            kwargs["object"] = object

        if target is not None:
            kwargs["target_type"] = target.entity_type
            if sa.inspect(target).deleted:
                kwargs["target_id"] = target.id
            else:
                kwargs["target"] = target

        entry = ActivityEntry(**kwargs)
        entry.object_type = object.entity_type
        session.add(entry)

    @staticmethod
    def entries_for_actor(actor: User, limit: int = 50) -> list[ActivityEntry]:
        return (
            ActivityEntry.query.filter(ActivityEntry.actor == actor).limit(limit).all()
        )



from __future__ import annotations

from pathlib import Path
from typing import Any, Dict

import jinja2
from flask import Flask, current_app
from flask.helpers import locked_cached_property
from flask.templating import Environment
from flask_babel import get_locale as babel_get_locale
from jinja2.loaders import ChoiceLoader, PackageLoader
from sqlalchemy.orm.attributes import NEVER_SET, NO_VALUE

import abilian.core.util
import abilian.i18n
from abilian.web import csrf
from abilian.web.filters import init_filters
from abilian.web.util import url_for
from abilian.web.views.images import user_photo_url


class JinjaManagerMixin(Flask):
    def __init__(self):
        self._jinja_loaders = []

    #
    # Templating and context injection setup
    #
    def create_jinja_environment(self) -> Environment:
        env = Flask.create_jinja_environment(self)
        env.globals.update(
            app=current_app,
            csrf=csrf,
            get_locale=babel_get_locale,
            local_dt=abilian.core.util.local_dt,
            _n=abilian.i18n._n,
            url_for=url_for,
            user_photo_url=user_photo_url,
            NO_VALUE=NO_VALUE,
            NEVER_SET=NEVER_SET,
        )
        init_filters(env)
        return env

    @locked_cached_property
    def jinja_options(self) -> dict[str, Any]:
        options = dict(Flask.jinja_options)

        jinja_exts = options.setdefault("extensions", [])
        ext = "abilian.core.extensions.jinjaext.DeferredJSExtension"
        if ext not in jinja_exts:
            jinja_exts.append(ext)

        if "bytecode_cache" not in options:
            cache_dir = Path(self.instance_path, "cache", "jinja")
            if not cache_dir.exists():
                cache_dir.mkdir(0o775, parents=True)

            options["bytecode_cache"] = jinja2.FileSystemBytecodeCache(
                str(cache_dir), "%s.cache"
            )

        if self.debug and self.config.get("TEMPLATE_DEBUG", False):
            options["undefined"] = jinja2.StrictUndefined
        return options

    def register_jinja_loaders(self, *loaders: PackageLoader):
        """Register one or many `jinja2.Loader` instances for templates lookup.

        During application initialization plugins can register a loader so that
        their templates are available to jinja2 renderer.

        Order of registration matters: last registered is first looked up (after
        standard Flask lookup in app template folder). This allows a plugin to
        override templates provided by others, or by base application. The
        application can override any template from any plugins from its template
        folder (See `Flask.Application.template_folder`).

        :raise: `ValueError` if a template has already been rendered
        """
        if not hasattr(self, "_jinja_loaders"):
            raise ValueError(
                "Cannot register new jinja loaders after first template rendered"
            )

        self._jinja_loaders.extend(loaders)

    @locked_cached_property
    def jinja_loader(self) -> ChoiceLoader:
        """Search templates in custom app templates dir (default Flask
        behaviour), fallback on abilian templates."""
        loaders = self._jinja_loaders
        del self._jinja_loaders
        loaders.append(Flask.jinja_loader.func(self))
        loaders.reverse()
        return jinja2.ChoiceLoader(loaders)



from __future__ import annotations

import datetime
from textwrap import dedent
from typing import Iterator

import html5lib
from flask import Flask
from flask_babel import Babel
from jinja2 import Environment
from pytest import fixture
from pytz import timezone, utc

from .. import filters

env = Environment()
filters.init_filters(env)


def en_locale() -> str:
    return "en"


def user_tz() -> str:
    # This one is GMT+8 and has no DST (tests should pass any time in year)
    return "Asia/Hong_Kong"


USER_TZ = timezone(user_tz())


def test_labelize():
    labelize = filters.labelize
    assert labelize("test_case") == "Test Case"


def test_filesize():
    filesize = filters.filesize
    assert str(filesize("100")) == "100&nbsp;B"
    assert str(filesize(100)) == "100&nbsp;B"
    assert str(filesize(1000)) == "1.0&nbsp;kB"
    assert str(filesize(1100)) == "1.1&nbsp;kB"
    assert str(filesize(10000)) == "10&nbsp;kB"
    assert str(filesize(1100100)) == "1.1&nbsp;MB"
    assert str(filesize(10000000)) == "10&nbsp;MB"
    assert str(filesize(1100100000)) == "1.1&nbsp;GB"
    assert str(filesize(100000000000)) == "100&nbsp;GB"


def test_roughsize():
    roughsize = filters.roughsize
    assert "6" == roughsize(6)
    assert "15" == roughsize(15)
    assert "130+" == roughsize(134)
    assert "10+" == roughsize(15, above=10)
    assert "55+" == roughsize(57, mod=5)


def test_abbrev():
    abbrev = filters.abbrev
    assert "test" == abbrev("test", 20)
    assert "Longer test...e truncated" == abbrev(
        "Longer test. it should be truncated", 25
    )


def test_linkify():
    tmpl = env.from_string('{{ "http://test.example.com"|linkify}}')
    rendered = tmpl.render()
    el = html5lib.parseFragment(rendered)
    assert len(list(el)) == 1

    el1 = el[0]
    assert el1.tag == "{http://www.w3.org/1999/xhtml}a"
    assert el1.text == "http://test.example.com"
    assert sorted(el1.items()) == [
        ("href", "http://test.example.com"),
        ("rel", "nofollow"),
    ]


def test_nl2br():
    tmpl = env.from_string(
        '{{ "first line\nsecond line\n\n  third, indented" | nl2br }}'
    )
    assert (
        tmpl.render()
        == "first line<br />\nsecond line<br />\n<br />\n  third, indented"
    )


def test_paragraphs():
    markdown_text = dedent(
        """\
        {{ "First paragraph
        some text
        with line return

        Second paragraph
        ... lorem

        Last one - a single line" | paragraphs }}
        """
    )
    tmpl = env.from_string(markdown_text)

    expected = dedent(
        """\
        <p>First paragraph<br />
        some text<br />
        with line return</p>

        <p>Second paragraph<br />
        ... lorem</p>

        <p>Last one - a single line</p>
        """
    )
    assert tmpl.render().strip() == expected.strip()


@fixture
def app() -> Iterator[Flask]:
    app = Flask(__name__)
    babel = Babel(app, default_locale="fr", default_timezone=USER_TZ)
    babel.localeselector(en_locale)
    babel.timezoneselector(user_tz)
    with app.app_context():
        yield app


def test_date_age(app: Flask):
    date_age = filters.date_age
    now = datetime.datetime(2012, 6, 10, 10, 10, 10, tzinfo=utc)
    assert date_age(None) == ""

    dt = datetime.datetime(2012, 6, 10, 10, 10, 0, tzinfo=utc)
    assert date_age(dt, now) == "2012-06-10 18:10 (1 minute ago)"

    dt = datetime.datetime(2012, 6, 10, 10, 8, 10, tzinfo=utc)
    assert date_age(dt, now) == "2012-06-10 18:08 (2 minutes ago)"

    dt = datetime.datetime(2012, 6, 10, 8, 30, 10, tzinfo=utc)
    assert date_age(dt, now) == "2012-06-10 16:30 (2 hours ago)"

    # for coverage: test when using default parameter now=None
    # dt_patcher = mock.patch.object(
    #     datetime, "datetime", mock.Mock(wraps=datetime.datetime)
    # )
    # with dt_patcher as mocked:
    #     mocked.utcnow.return_value = now
    #     assert date_age(dt) == "2012-06-10 16:30 (2 hours ago)"


def test_age(app: Flask):
    age = filters.age
    now = datetime.datetime(2012, 6, 10, 10, 10, 10, tzinfo=utc)
    d1m = datetime.datetime(2012, 6, 10, 10, 10, 0, tzinfo=utc)
    d3w = datetime.datetime(2012, 5, 18, 8, 0, 0, tzinfo=utc)
    d2011 = datetime.datetime(2011, 9, 4, 12, 12, 0, tzinfo=utc)

    # default parameters
    assert age(None) == ""
    assert age(d1m, now) == "1 minute ago"
    assert age(d3w, now) == "3 weeks ago"

    # with direction
    assert age(d1m, now, add_direction=False) == "1 minute"
    assert age(d3w, now, add_direction=False) == "3 weeks"

    # with date_threshold
    assert age(d1m, now, date_threshold="day") == "1 minute ago"
    # same year: 2012 not shown
    assert age(d3w, now, date_threshold="day") == "May 18, 4:00 PM"
    # different year: 2011 shown
    assert age(d2011, now, date_threshold="day") == "September 4, 2011, 8:12 PM"

    # using default parameter now=None
    # dt_patcher = mock.patch.object(
    #     datetime, "datetime", mock.Mock(wraps=datetime.datetime)
    # )
    # with dt_patcher as mocked:
    #     mocked.utcnow.return_value = now
    #     assert age(d1m) == "1 minute ago"



""""""
from __future__ import annotations

import imghdr
from io import BytesIO

import babel.dates
import PIL.Image
from flask import current_app, flash, g, redirect, render_template, request, url_for
from werkzeug.exceptions import InternalServerError
from wtforms.fields import StringField
from wtforms.validators import ValidationError

from abilian.core.extensions import db
from abilian.i18n import _, _l, get_default_locale
from abilian.services.preferences.panel import PreferencePanel
from abilian.web import csrf
from abilian.web.forms import Form, fields, required, widgets
from abilian.web.forms.fields import FileField


class UserPreferencesForm(Form):

    password = StringField(
        _l("New Password"), widget=widgets.PasswordInput(autocomplete="off")
    )
    confirm_password = StringField(
        _l("Confirm new password"), widget=widgets.PasswordInput(autocomplete="off")
    )

    photo = fields.FileField(
        label=_l("Photo"), widget=widgets.ImageInput(width=55, height=55)
    )

    locale = fields.LocaleSelectField(
        label=_l("Preferred Language"),
        validators=[required()],
        default=lambda: get_default_locale(),
    )

    timezone = fields.TimezoneField(
        label=_l("Time zone"), validators=(required(),), default=babel.dates.LOCALTZ
    )

    def validate_password(self, field: StringField):
        pwd = field.data
        confirmed = self["confirm_password"].data

        if pwd != confirmed:
            raise ValidationError(
                _(
                    "Passwords differ. Ensure you have typed same password "
                    'in both "password" field and "confirm password" field.'
                )
            )

    def validate_photo(self, field: FileField):
        data = request.form.get(field.name)
        if not data:
            return

        data = field.data
        filename = data.filename
        valid = any(filename.lower().endswith(ext) for ext in (".png", ".jpg", ".jpeg"))

        if not valid:
            raise ValidationError(_("Only PNG or JPG image files are accepted"))

        img_type = imghdr.what("ignored", data.read())

        if img_type not in ("png", "jpeg"):
            raise ValidationError(_("Only PNG or JPG image files are accepted"))

        data.seek(0)
        try:
            # check this is actually an image file
            im = PIL.Image.open(data)
            im.load()
        except Exception:
            raise ValidationError(_("Could not decode image file"))

        # convert to jpeg
        # FIXME: better do this at model level?
        jpeg = BytesIO()
        im.convert("RGB").save(jpeg, "JPEG")
        field.data = jpeg.getvalue()


class UserPreferencesPanel(PreferencePanel):
    id = "user"
    label = _l("About me")

    def is_accessible(self):
        return True

    def get(self):
        # Manual security check, should be done by the framework instead.
        if not self.is_accessible():
            raise InternalServerError()

        data = {}
        photo = g.user.photo
        if photo:
            # subclass str/bytes to set additional 'url' attribute

            photo = type(
                "Photo",
                (bytes,),
                {"object": photo, "url": url_for("users.photo", user_id=g.user.id)},
            )
            data["photo"] = photo

        form = UserPreferencesForm(obj=g.user, formdata=None, prefix=self.id, **data)
        if form["locale"].data is None:
            form["locale"].data = get_default_locale()

        ctx = {"form": form, "title": self.label}
        return render_template("preferences/user.html", **ctx)

    @csrf.support_graceful_failure
    def post(self):
        # Manual security check, should be done by the framework instead.
        if not self.is_accessible():
            raise InternalServerError()

        if request.form["_action"] == "cancel":
            return redirect(url_for(".user"))

        form = UserPreferencesForm(request.form, prefix=self.id)

        if form.validate():
            if request.csrf_failed:
                current_app.extensions["csrf-handler"].flash_csrf_failed_message()
                return render_template("preferences/user.html", form=form)

            del form.confirm_password
            if form.password.data:
                g.user.set_password(form.password.data)
                flash(_("Password changed"), "success")
            del form.password

            form.populate_obj(g.user)

            db.session.commit()
            flash(_("Preferences saved."), "info")
            return redirect(url_for(".user"))
        else:
            return render_template("preferences/user.html", form=form)
